/*
  h_file.c   Nov98  JS  version 0.3
  
  this is a file for handling Htmp-File (tmp_hxy.000) 
  reasons and goals for this file:
  - 2GB file breaking => use LFS (64bit) or more blocks
  - error handling (perror nach printf???) Jun99
  - open/close
  - hidden variables (encapsulating of file-handling)
  - better reading of main-file
  Ideas:
  - chksum for tmp-file, check during read?
  - vvv==1 ? measure disk performance
  - compressed tmp_hxy.XXX ? (example: 1GB => 336MB, zip=17m  unzip=3m read=1m)
                        thor  gzip -c1 1GB => 640MB, zip=10m  unzip=3m
                        thor  gzip -c9 1GB => 579MB, zip=125m unzip=3m 
                        thor  gzip -c4 1GB => 550MB, zip=14m  unzip=3m
  - see pipe.c!!! (h_file,fopen,fclose,... & fn,pipe_ropen,... kapseln
  - repeat H*v if error!?
  - see also utils/vbuf.c for buffering/speed/IO
  
  this file is still changing Nov98,Apr99,Jun99,Nov99
  history:
   v2.17 - array of thxy -to- thxy of arrays
   v2.0.1 - access to scplx only via Macros M_d2f(), M_f2d()
   v2.0.0 - file-split removed (see history)
   03.02.00 - bug cplx rr => scplx rr
   25.11.99 - bug in my_fread/write if file > 1MB beseitigt
   07.01.00 - bug auf MIPS beseitigt (high-low Format!!!)
               Vorsicht bei type-Umwandlungen (high-low/low-high format)
   ToDo: remove pipe.c (simpler), add mem-only-version (mmap?) /dev/shm???

*/
#include <stdio.h>
#include <stdlib.h>
/* #include <errno.h> 2015-12-07 not needed? t100.gpu02 no errno */
#include <string.h>	/* strerror */
#include "defs.h"
#include "spins.h"
#include "h_file.h"
#include "pipe.h"
#include "block.h"
#if 1
#include <unistd.h>		/* test this with configure! */
#define SLEEP(x) sleep(x)
#else
#define SLEEP(x) {}
#endif

/*
 * v2.30 can have different numbers of I/O threads and compute threads
 */

#ifndef HRMAX
#if HRMAX>0
#error "HRMAX!=0 if not defined?" 
#endif
#endif
#if HRMAX>0	/* table for 16bit index to Hamilton-entries */
 int  hr_len[B_NUM];    /* number of table entries */
 cplx hr[B_NUM][HRMAX];	/* table of matrix values, one table per thread */
 /* better use realloc for dyn. pt_n ? */
#endif

/* define the basename for the H-files htmpBBBB.PPP
 * where BBB is the IO-thread-number (block)
 * and PPP the paramter index (ToDo)
 */
#define Htmp "./tmp/htmp"	/* ./tmp can be a link to /tmp/$USER */

/* public variables */

/* privat variables */
char  h_name[B_NUM][40]; /* file-name B_NUM=(ToDo: <=)B_NUM files */
int   h_error=0;         /* set if H could not be written */
int   h_state[B_NUM];    /* 0=closed, 1=ropen, 2=wopen */
int   h_open[B_NUM];    /* number cthreads have files opened 0..B_NUM/F_NUM-1 */
static int out_vvv=0;   /* number of debug outputs */
/* 2013-04: allow less but interleaved block files on shared slow disks  
 *   disk speed may vary between 100% to less than 10% for parallel files
 *  but each task must send at least zero block until it gets ready_for_close
 *  ToDo: alternatively synchronize blockwrites using ring-token
 *        num_IO_synced_threads instead of threads per file
 *  ring_tokens bit0=ORed_some_thrads_had_IO, bit1=1(no_IOs_last_round=close)
 */
/* int threads_per_file=1; */ /* default: 1 = 1 file/task, pt_n = 1 file/node */

/* my comfortable erroroutput for misterious errors */
int my_ferror(const char *msg){ /* JS-Mar11 add const */
  perror(msg);	/* errno_msg => stderr */
  SLEEP(1);
  return 0;
}


/* including error-handling (public???) */
size_t my_fread( char *ptr, size_t size, size_t nmemb, int block){
  size_t n_readed=0,nf,ni,tries=0; /* sgi-CC can not add void* */
  nmemb*=size;
  if (block>=B_NUM) ERR(9999);
  if( nmemb )
  while(n_readed< size /* nmemb */ ){ /* mind. 1 element */
    /* some clibs do alloc a buffer for disk_read which cost time if it is to big */
    ni=nmemb-n_readed;
    /* v2.29 add tries>0 because I forgot whats the splitting is good for */
    if(tries>0 && ni>262144) ni=262144;	/* 2^n=256kB pieces, why? */
    nf=pipe_read(ptr,(int)ni,block);
    if(nf<=0){ E3("read=%d ni=%d block=%d ",(int)nf,(int)ni,block);
               my_ferror("read");if(tries++<5)continue;break; }
    n_readed+=nf; ptr+=nf; tries=0;
  } /* while */
  return n_readed/size;
}

/* 
   write large objects in small 1e6-pieces,
    because fwrite (probably) need lot of memory
   size should be Hsize = multiple of sector size ! (4k and bigger)
 */
size_t my_fwrite( void *ptr, size_t size, size_t nmemb, int block){
  size_t n_written=0, nf, ni, tries=0;
  char *p=(char *)ptr;	/* sgi-CC can not add void* */
  if (mpi_n*pt_n<=8 && (vvv&16) && out_vvv<1000) { out_vvv++;
    printf("\nvvv&16 fwrite task=%3d size=%d*%d %p err=%d ",
           mpi_i*pt_n+block, (int)size, (int)nmemb, ptr, h_error);
    fflush(stdout); }
  if (block>=B_NUM) ERR(9999); /* cant be! */
  if (h_state[block]!=2) ERR(9999); /* file not open */
  nmemb*=size;
  if (h_error) return 0;  /* dont try if something went wrong on I/O */
  while (n_written<nmemb) {
    ni=nmemb-n_written;
    /* v2.29 I forgot whats splitting good for, try to speedup */
#if STOREH >= 2 && STOREH <= 4   /* dont split malloc blocks, JS0904 */
    if (tries>0 && ni>262144) ni=262144; /* 256kB pieces, avoiding OOM? */
#endif
    nf=pipe_write(p,(int)ni,block);
    n_written+=nf; p+=nf;
    if(nf<ni){
      ERR(620);
      if (h_error) break;  /* dont try 3 times on old errors */
      tries++;
#if STOREH != 0    /* dont repeat on low mem condition */
      sleep(1<<(2*tries)); /* time to relax 1,4,16,64s */
      if (tries<3) continue; /* try 3 times */
#endif
      h_error=1; break;    /* write finally failed */
    }
    tries=0;
  } /* end while */
  /* if (vvv&16) printf(" %d/%d -",n_written,size); */
  return n_written/size;
}

/* --- end of file handler --- */

/* --------- interface between compute blocks and H storage ------ */

/* open for reading */
int h_ropen(int block){ /* 0..B_NUM-1, block > B_NUM possible */
  if (mpi_n*pt_n<=8 && (vvv&16)) /* max 8 Lines debugging or no output */
    printf("\nvvv&16 h_ropen  %3d", mpi_i*pt_n+block);
  if (block<0 || block>=pt_n) ERR(9999); /* for safety */
  if (++h_open[block%B_NUM]==1) { /* open stream for first block */
    if (mpi_n*pt_n<=8 && (vvv&16)) printf(" open %s", h_name[block%B_NUM] );
    if ( pipe_ropen(h_name[block],block) ){
      E1("fopen_%s",h_name[block]);return 1; }
  }
  b_bnz[block]=HBLen; /* force h_read on first h_get */
  if (h_state[block]!=0) { F1("h_ropen(%d): file was open", block); }
  h_state[block]=1;  /* opened for read */
  return 0;
}

/* open for writing */
int h_wopen(int block){  /* 0..B_NUM-1, block > B_NUM possible */
  if (block<0 || block>=pt_n) ERR(9999); /* for safety */
  /* JS-Apr2011 for better diff output, 0000...%04d mpi_n*pt_n-1 replaced */
  if(vvv && mpi_i*pt_n+block==0)
    printf(" write/store " Htmp "XXXX STOREH=%d(0=mem,2=dsk,...)",
           STOREH);  /* Hsize = size.struct...[HBlen] */
  if (mpi_n*pt_n<=8 && (vvv&16)) /* maximum debug info for 8 threads */
     printf("\nvvv&16 h_wopen  task=%3d open=%d",
             mpi_i*pt_n+block, h_open[block%B_NUM]);
  if (++h_open[block%B_NUM]==1) { /* open stream for first block */
    if (mpi_n*pt_n<=8 && (vvv&16))
       printf(" pipe_wopen task=%2d", block+mpi_i*pt_n);
    if ( pipe_wopen(h_name[block%B_NUM],block%B_NUM) ) { 
      E1("wopen %s",h_name[block%B_NUM]); h_error=1; return 1; }
  }
  b_bnz[block]=0;
  b_bn1[block]=0;
  if (h_state[block]!=0) ERR(9999); /* file not open */
  h_state[block]=2;  /* opened for write */
  if (mpi_n*pt_n<=8 && (vvv&16)) printf(" ok");
  return 0;
}

int h_rclose(int block) {
  if (mpi_n*pt_n<=8 && (vvv&16)) /* max 8 lines output */
    printf("\nvvv&16 h_rclose %3d open=%d",
            mpi_i*pt_n+block, h_open[block%B_NUM]);
  if (h_state[block]!=1) ERR(9999); /* file not open */
  h_state[block]=0;                 /* mark as closed */
  if (--h_open[block%B_NUM]==0) /* close stream if last block flushed */
    pipe_close(block%B_NUM);
  return 0;
}

/* --- read from Htmp-file, ret: 0=ok --- 
 * read from disk into H-buffer, struct thxy = xy[HBLen], r[HBLen](or idx+para)
 *  ToDo: get special block from stream of random(?) stored blocks 
 *  search dynamic buffer, if not found extend buffer until block found
 *  max. B_NUM/B_NUM (* PARAM)?
 * this buffer is written by B_NUM IO-threads
 *   and read by B_NUM compute threads (v2.26 deadlocks?)
 * IO-thread=file blocks=compute-threads
 *                          v-blocks IO=0-Buffer? ToDo
 *    0:          0,1,4,0,5,0,1,4 (compute block 0 was faster than the others?)
 *    1:          2,3,6,7,2,3,6,7
 */
 
#if defined(STOREH) && STOREH==0
void *pmnext(int block); /* pipe.c */
#endif
/* privat */
/* read special block, !!! ToDo: read next block, caller must handle that
 *   read from stream[blk] or memory[blk] to tbh (i.e. b_hbuf[block])
 *   for STOREH==0 we simply return tbh as pointer to next H_block 
 */
struct thxy *h_read(struct thxy *tbh, int block){
#if defined(STOREH) && STOREH==0
  /* pipe_read((char *)tbh, Hsize, block); # old v2.35 */
  return ((struct thxy *)pmnext(block)); /* v2.36 Jul08 */
#else
  if (vvv&16 && out_vvv<1000)  /* print before, because read may hang */
    printf("\nvvv&16 h_read %3d state=%d", mpi_i*pt_n+block, h_state[block]);
  if (h_state[block]!=1) ERR(9999); /* file not open */
  if (my_fread((char *)tbh/*&b_hbuf[block]*/,Hsize,1,block%B_NUM)!=1){
     E0("h_read"); ERR(9622); return tbh;
   }
  if (vvv&16 && out_vvv<1000) { out_vvv++;  /* move to h_read */
    printf("\nvvv&16 h_read %3d state=%d"
           " bj %2d j %6ld r %3d jf %d ...",
          mpi_i*pt_n+block, h_state[block],
           (int) tbh->/*b_hbuf[block].*/by[ 0 ],
          (long) tbh->/*b_hbuf[block].*/xy[ 0 ]&~XY_FLAG,
           (int) tbh->/*b_hbuf[block].*/rr[ 0 ],
          (int) ((tbh->/*b_hbuf[block].*/xy[ 0 ] & XY_FLAG)?1:0)); }
  return tbh;  /* return same buffer, data copied in */
#endif
}


/* ------------------------------------------------------------------ *
 * h_put: write matrix-element and its column (line encoded)          *
 * h_get: read  matrix-element and its column (line+1 if column>n1)   *
 * ------------------------------------------------------------------ */
#define HR_SIZE   sizeof(HR_TYPE)
#define XY_SIZE   sizeof(XY_TYPE)  


#if HRMAX>0
/* seperate error routine, only loaded to CPU-cache in error-case */
void err_fulltable(){
    int i, j;
    /* output maximum on information, what was went wrong? */
    printf("\nerr_fulltable: --debug--\n");
    for (i=0;i<pt_n;i++) if (hr_len[i]>=HRMAX-4) {
      printf(" Hblock=[%d.%d] b_hnz=%9ld hr_len=%4d hr[]=",
             mpi_i, i, b_hnz[i], hr_len[i]);
      for (j=0;j<hr_len[i] && j<100;j++) printf(" %g",D(hr[i][j]));
      if (j>=100) printf(" ...");
      printf("\n");
    }
    ERR(9100);
}
#endif

/* h_put fills a write buffer, this fkt writes the buffer out!
 * h_wclose is only called to close the file after writing
 * block=1..B_NUM-1 can be equal and greater than F_NUM to flush all buffers
 * called by storeh for every thread + from storeh2 thread0
 */
int h_wclose(int block){  /* store in file */
  /* JS0904: returning on error causes more errors if H partly stored */
  /* if(h_error) return 1; */ /* bad, if partly stored */
  if (vvv&16)
    printf("\nvvv&16 h_wclose %3d open=%d", mpi_i*pt_n+block, h_open[block%B_NUM]);
  if (h_state[block]!=2) ERR(9999); /* file not open */
  h_state[block]=0; /* mark as closed */
  if (--h_open[block%B_NUM]==0) { /* close stream if last block flushed */
    if (vvv&16) printf(" pipe_close %3d", (mpi_i*pt_n+block)%B_NUM);
    pipe_close(block%B_NUM); }
  return 0;
}

/* h_put and h_get moved to spins.c
 * to make the call shorter and allow inlining => no speedup Jan07 */

/* initialize H-storage, called by spins.c storeh*(), thread0 only */
void ini_thxy(ULong n1){ /* alloc */
  static int num_ini=0;   /* malloc only once */
  int i1,i2,j; ULong i; XY_TYPE xy; struct thxy t1;
  if (!num_ini) {
     // if (!b_hbuf) b_hbuf=(struct thxy *)malloc(pt_n*sizeof(struct thxy));
     // if (!b_hbuf) ERR(9999);
  } num_ini++;
  for(j=0;j<pt_n;j++){
    h_state[j]=0; /* default 0=close */
#if HRMAX>0
    hr_len[j]=0; /* table for index-to-matrix value */
#endif
  }
  for(i1=0;i1<B_NUM;i1++){
    h_open[i1]=0;               /* no bocks open */
    h_name[i1][0]=0;		/* file-name ToDo: set name here! */
    sprintf(h_name[i1],Htmp "%04u.%03u",(int)i1+mpi_i*B_NUM,1);
    /* ToDo: set by user? spread over disks, better use links */
/*    for (i2=0;i2<pt_n;i2++)
      hblockbufs[i1][i2].block=-1;
*/ /* mark as free, ToDo */
  }
  /* measure: i1 = used bytes for xy + XY_FLAG,
   *          i2 = sizeof xy in bytes (ToDo: check sense)
   * we store 0..node_n1 (not n1) to .xy, less bits
   * i must be unsigned to avoid endless loops if first bit set
   */
  for (j =0,i=2*     n1;i!=0; j++, i>>=1); /* n1 bits + FLAG (not needed) */
  for (i1=0,i=2*node_n1;i!=0;i1++, i>>=1); /* node_n1 bits + FLAG (Jan09) */
  for (i2=0,xy=1;      xy!=0;i2++,xy<<=1);
  if (vvv && !mpi_i) 
    printf("\n ini_thxy.bits:  thxy.xy %d  BY %d  HR %d  node_n1,n1 %d %d",
       i2,  8*(int)sizeof(BY_TYPE), 8*(int)sizeof(HR_TYPE), i1, j);
  if (  i1> i2) F0("xy to small, enlarge h_file.h:XY_TYPE or add mpi-tasks");
  if (2*i1<=i2 && i1>20 && !mpi_i) /* dont warn on play systems */
    printf("\nWarning: reduce h_file.h:XY_TYPE to spare memory.");
  /* h_file.h struct thxy XY_TYPE */
  j=sizeof(int); /* avoid compiler warnings for if(const1==const2) */
  if (j!=sizeof(t1.xy[0]) && !mpi_i) E0("sizeof int!=thxz.xy"); /* ToDo: why? */
  if (vvv && !mpi_i) 
    printf("\n");
}

#define HR_FILE "tmp/hr_s%04d.dat"  /* ToDo: mpi.001 etc. */

/* checkpoint Hstorage, write restore data to HR_FILE
 ToDo: writing to mpi_n local files(?) (distributed or shared medium)
 ToDo: store a unique hash to validate having correct files
   get_max_hr MPI_IO_open + seek(mpi_i*max_hr*hrsize) + save (shared file?)
   suspend/resume only if mpi_n not changed !? 
*/
void h_checkpoint(int vvv){
  char s1[80]="dummy_checkpoint";
  int i, hrmax;
  FILE *f1=NULL;
  /* if( save_mode & 1 ) --- small amount of data! save every time */
#if STOREH != 0  /* makes no sence for H in memory, v2.32 */
  /* save data before possible crash */
  sprintf(s1,HR_FILE,mpi_i);
  f1=fopen(s1,"wb"); if (!f1) E1("error fopen %s",s1);
#endif
  hrmax=0; /* max hr-table_length */
  for (i=0;i<pt_n;i++) {  /* all pthreads */
#if HRMAX>0
    if (hr_len[i]>hrmax) hrmax=hr_len[i];
    /* ToDo: better output num threads*mpi + hr_min/max */
    // ToDo: MPI_.... min,max output? loop !?
    if ((!mpi_i) && i==0)
      printf("\nDBG: h_checkpoint.w thread=0..%d ...\n", pt_n);
    if (mpi_i==1) puts("\nWARNING: h_checkpoint not ready for MPI code");
    if (vvv && mpi_i==0 && i<=8)
      printf(((i==0)?" hr[] len: %d":" %d"),hr_len[i]);
    if (f1){
      if (fwrite(&hr_len[i],sizeof(hr_len[i]), 1, f1)!=1
       || fwrite(&hr[i][0], sizeof(hr[i][0]), hr_len[i], f1)
                                    !=(size_t)hr_len[i]) 
        E1("h_checkpoint.fwrite pt_i=%d", i); /* hr_len==0 ??? */
    }
#endif
    if (f1) /* 2012-01 ToDo: within pt-loop?, 2012-04 add error handling */
      if (fwrite(b_hnz, sizeof(b_hnz[0]), pt_n, f1)!=(size_t)pt_n
       || fwrite(b_snz, sizeof(b_snz[0]), pt_n, f1)!=(size_t)pt_n
       || fwrite(b_sn1, sizeof(b_sn1[0]), pt_n, f1)!=(size_t)pt_n)
        E0("h_checkkpoint");
  } // i= all pthreads
#ifdef CONFIG_MPI
  MPI_Allreduce(&hrmax,&i,1,MPI_INT,MPI_MAX,MPI_COMM_WORLD); hrmax=i;
#endif
  if (f1) fclose(f1);
  if (vvv && !mpi_i) { /* to minimize hr_bits in H, we have to know hr_num */
    printf("\n h_checkpoint.w: max hr_num= %5d (entries/block) %s..\n",
           hrmax, s1);
  }
}

/* restore H-storage, after h_checkpoint, same nodes (local disks) */
/* load data after crash, defekt node must be replaced? better recalculate! */
void h_restore(int vvv){
  int i, hrmax=0;
  FILE *f1=NULL;
  if (!mpi_i) fputs(" h_restore(chkpt>=4,STOREH!=0 only) ...",stdout);
#if STOREH != 0  /* makes no sence for H in memory, v2.32 */
  char s1[80];
  sprintf(s1,HR_FILE,mpi_i);
  f1=fopen(s1,"rb");if(!f1)E0("error fopen");
#endif
#if HRMAX>0
  if (f1)
  for(hrmax=i=0;i<pt_n;i++){
    if (vvv) printf(" load hr[%d]=",i);
    if (fread(&hr_len[i],sizeof(hr_len[i]),1,f1)!=1  // ToDo: correct? 2012
    ||  fread(&hr[i][0], sizeof(hr[i][0]), hr_len[i],f1)!=(size_t)hr_len[i])
      F0("h_restore.fread.HR");
    if ( hr_len[i]>hrmax ) hrmax=hr_len[i];
    if (vvv) printf("%d",hr_len[i]);
  }
#endif
  if (f1)
    if (fread(b_hnz, sizeof(b_hnz[0]), pt_n, f1)!=(size_t)pt_n
    ||  fread(b_snz, sizeof(b_snz[0]), pt_n, f1)!=(size_t)pt_n
    ||  fread(b_sn1, sizeof(b_sn1[0]), pt_n, f1)!=(size_t)pt_n)
     F0("h_restore.fread");
  if (f1) fclose(f1);
#ifdef CONFIG_MPI /* ToDo: bad code !? */
  MPI_Allreduce(&hrmax,&i,1,MPI_INT,MPI_MAX,MPI_COMM_WORLD); hrmax=i;
#endif
  if (vvv && !mpi_i)
    printf(" h_restore: max hr_num= %5d (entries/block)\n",hrmax);
}
/* ---------------------------------------------------------- */
